<?lassoscript

	// [Encode_QuotedPrintable] - Encodes email bodies using quoted printable algorithm.
	//
	// -Data = (string, the text to be encoded)
	// -Charset = (optional, desired character set after decoding, defaults to 'utf-8')
	//
	// The tag encodes any characters within the ASCII range 0x20 to 0x7e except 0x5f and 0x3d
	// (that is all standard ASCII characters from space to tilde except for underscore and the
	// equal sign).  Lines are automaticaly truncated to less than 76 characters and joined with
	// \r\n.
	//

	define_tag: 'quotedprintable',
		-priority = 'replace',
		-required = 'data',
		-optional = 'charset',
		-namespace = 'encode_';

		if: !#data->size;
			return: string;
		/if;

		if: !(local_defined: 'charset');
			local: 'charset' = 'utf-8';
		/if;

		fail_if: !(string_validcharset: #charset), -9956, 'Expected a valid character set but got ' + #charset + '.';

		local:
			'braw' = bytes,
			'qped' = bytes,
			'linecounter' = 0,
			't' = null;

		#braw->(importstring: (string_replaceregexp: -find = '\r\n|\r|\n', -replace = '\r\n', #data), #charset);

		if: (#data >> '\r\n.') || (#data >> '\r\nF') || (#data >> '\r\nf') || (string_findregexp: -find = '[^\\x09\\x0A\\x0D\\x20-\\x3C\\x3E-\\x5E\\x60-\\x7E]', #data)->size;
			// Encode text if a line starts with .fF or if it contains out-of-range ASCII characters

			iterate: #braw, (local: 'i');
				if:
					(#i == 0x9) || // \t
					(#i == 0xD) || // \r
					(#i == 0xA) || // \n
					(
						(#i >= 0x20) &&
						(#i <= 0x7E) &&
						(#i != 0x5F) && // _
						(#i != 0x3D)    // =
					);

					if: (#i == 0x0A) || (#i == 0x0D);
						#linecounter = 0;
					else: (#linecounter++ >= 75);
						#qped += '=\r\n';
						#linecounter = 1;
					/if;

					if: (#linecounter == 1) && (#i == 0x2E);
						// Encode starting .
						#qped += '=2E';
						#linecounter += 3;
					else: (#linecounter == 1) && (#i == 0x46);
						// Encode starting F
						#qped += '=46';
						#linecounter += 3;
					else: (#linecounter == 1) && (#i == 0x66);
						// Encode starting f
						#qped += '=66';
						#linecounter += 3;
					else;
						#qped->(import8bits: #i);
					/if;

				else;

					#linecounter += 3;
					if: (#i == 0x0A) || (#i == 0x0D);
						#linecounter = 0;
					else: #linecounter >= 74;
						#qped += '=\r\n';
						#linecounter = 1;
					/if;

					#t = (string(#i->setformat(-hexadecimal=true) &)->removeleading('0x') & uppercase &);
					if: #t->size == 1;
						#t = '0' #t;
					/if;

					#qped += '=' #t;
				/if;
			/iterate;

		else;
			iterate: #braw, (local: 'i');
				if: (#i == 0x0A) || (#i == 0x0D);
					#linecounter = 0;
				else: (#linecounter++ >= 75);
					#qped += '=\r\n';
					#linecounter = 1;
				/if;

				if: (#linecounter == 1) && (#i == 0x2E) && ( (#braw->size == loop_count) || (#braw->(get: loop_count + 1) == 0x0D) );
					// Double up . on a line by itself
					#qped->(import8bits: #i);
					#qped->(import8bits: #i);
					#linecounter += 1;
				else;
					#qped->(import8bits: #i);
				/if;
			/iterate;
		/if;

		return: #qped->(exportstring: 'iso-8859-1');

	/define_tag;

	// [Decode_QuotedPrintable] - Decodes email bodies encoded using quoted printable algorithm.
	//
	// -Data = (string, the text to be decoded)
	// -Charset = (optional, character set of encoded text, defaults to 'utf-8')
	//
	
	define_tag: 'quotedprintable',
		-priority = 'replace',
		-required = 'data',
		-optional = 'charset',
		-namespace = 'decode_';

		if: !(local_defined: 'charset');
			local: '_charset' = 'utf-8';
		else: !(string_validcharset: #charset);
			local: '_charset' = 'iso-8859-1';
		else;
			local: '_charset' = #charset;
		/if;
		
		local: 'temp_output' = bytes;
		local: 'i' = 1;
		local: 'x' = '';
		while: (#i <= #data->length);
			if: (#data->(get: #i) == '=');
				local: 'entity' = #data->(substring: #i,3);
				if: ((string_replaceregexp: #entity, -find='=[0-9A-Z]{2}', -replace='') == '');
					#entity->(replace: '=','0x');
					#temp_output->(import8bits: (integer: #entity));
					#i += 3;
				else: ((string_replaceregexp: #entity, -find='=\\s\\s', -replace='') == '');
					#i += 3;
				else: ((string_replaceregexp: #entity, -find='=\\s', -replace='') == '');
					#i += 2;
				else;
					#temp_output += #data->(get: #i);
					#i += 1;
				/if;
			else;
				#temp_output += #data->(get: #i);
				#i += 1;
			/if;
		/while;
		
		local: 'output' = #temp_output->(exportstring: #_charset);
		return: #output;

	/define_tag;


	// [Encode_QHeader] - Encodes email headers using Q (quoted printable) algorithm
	//
	// -Name = (string, the name of the header to be encoded)
	// -Value = (string, the value of the header to be encoded)
	// -Charset = (optional, desired character set after decoding, defaults to 'utf-8')
	//
	// The tag encodes any characters within the ASCII range 0x20 to 0x7e except 0x5f and 0x3d
	// (that is all standard ASCII characters from space to tilde except for underscore and the
	// equal sign).  Lines are automaticaly truncated to less than 76 characters and joined with
	// \r\n.
	//
	// The output is =?charset?Q?text?= with any characters that require encoding changed to =xx 
	// entities within the text.  If no encoding is required then the charset and Q prefix are 
	// not included.  Individual words are encoded.
	//

	define_tag:'qheader',
		-Priority='Replace',
		-required='name', -type='string',
		-required='value', -type='string',
		-optional='charset',
		-namespace = 'encode_';

		local: '_name' = #name;
		#_name->trim;
		#_name->(removetrailing: ':');

		local: '_value' = #value;
		#_value->trim;
		#_value = (string_replaceregexp: #_value, -find='[\r\n\t]', -replace=' ');
		
		if: ( #_value->(beginswith: '=?') && (#_value >> '?Q?') && #_value->(endswith: '?=') );
			return: #_name + ': ' + #_value;
		/if;

		if: !(local_defined: 'charset');
			local: 'charset' = 'utf-8';
		/if;

		if: (#charset != 'none');
			fail_if: !(string_validcharset: #charset), -9956, 'Expected a valid character set but got ' + #charset + '.';
		/if;

		// encode all non-vanilla ascii characters as binary data
		// We encode anything other than %20 - %7e (std punc, 0-9, a-z, A-Z)
		local: 'prefix' = '';
		local: 'suffix' = '';
		if: (#charset != 'none') && (string_findregexp: -find = '[^\\x09\\x0A\\x0D\\x20-\\x7E]', #_value)->size;
			
			local: 'output' = bytes;

			local: 'array' = (string: #_value)->(split: ' ');
			iterate: #array, (local: 'temp');
				if: (loop_count > 1);
					#output += ' ';
				/if;
				if: (#temp->(beginswith: '<') && #temp->(endswith: '>')) || ((#temp >> '@') && (valid_email: #temp));
					#output += #temp;
				else;
					local:'input' = bytes;
					#input->(importString: #temp, #charset);
					iterate: #input, (local: 'd');
						if: ((#d >= 0x20) && (#d <= 0x7e) && (#d != 0x5f) && (#d != 0x3d));
							#output->(import8bits:#d);
						else;
							#d->(setformat: -hexadecimal=true);
							local: 's' = ((string_uppercase: #d) - '0x');
							if: #s->size == 1;#s = '0' + #s;/if;
							#output += '=' + #s;
						/if;
					/iterate;
				/if;
			/iterate;
			
			local: '_value' = #output->(exportString: 'iso-8859-1');
			local: 'prefix' = '=?' + #charset + '?Q?';
			local: 'suffix' = '?=';
		/if;

		// check line length, can not exceed 76 chars per line
		local: 'max_length' = 76 - (#_name->size + 2) - #prefix->size - #suffix->size;
		local: 'array' = (string_findregexp: #_value, -find='[^ ]{0,' + (#max_length - 6) + '}(?:[^ =]{2})? ?');
		local: '_value' = #_name + ': ';
		local: 'length' = #_value->size;
		local: 'size' = #array->size;
		iterate: #array, (local: 'temp');
			local: 'add' = '';
			if: (#temp >> '=');
				local: 'endfix' = '';
				if: #temp->(beginswith: '(') || #temp->(beginswith: '"');
					#add += #temp->(get: 1);
					#temp = #temp->(substring: 2);
				/if;
				if: #temp->(endswith: ')') || #temp->(endswith: '"');
					local: 'endfix' = #temp->(get: #temp->size);
					#temp = #temp->(substring: 1, #temp->size - 1);
				/if;
				#add += #prefix;
				if: (loop_count < #size);
					#add += #temp + '=20';
				else;
					#add += #temp;
				/if;
				#add += #suffix;
				#add += #endfix;
			else;
				#add += #temp;
			/if;
			#length += #add->size;
			if: #length <= #max_length;
				#_value += #add;
			else;
				#_value += '\r\n ' + #add;
				local: 'length' = #add->size + 1;
				local: 'max_length' = 76;
			/if;
		/iterate;		
		return: @#_value;

	/define_tag;


	// [Decode_QHeader] - Decodes email header values using Q (quoted printable) algorithm
	//
	// -Data = (string, the text to be decoded)
	//

	define_tag:'qheader',
		-Priority='Replace',
		-required='data', -type='string',
		-namespace = 'decode';

		local: 'output' = (string_replaceregexp: #data, -find='\r\n\\s+', -replace='');
		
		if: (#data !>> '=?') || (#data !>> '?q?') || (#data !>> '?=');
			return: #output;
		/if;
		
		local: 'output' = (string_replaceregexp: #output, -find='\\?=\\s+', -replace='?=');
		local: 'parse' = (string_findregexp: #output, -find='=\\?([a-zA-Z0-9_-]+?)\\?Q\\?(.*?)\\?=');

		if: #parse->size == 0;
			return: #output;
		/if;
		
		loop: -from=1, -to=#parse->size, -by=3;
			local: 'token' = #parse->(get: loop_count);
			local: 'charset' = #parse->(get: loop_count + 1);
			local: 'value' = #parse->(get: loop_count + 2);

			if: (string_validcharset: #charset);
				local: 'newvalue' = (decode_quotedprintable: (string_replace: #value, -find='_', -replace=' '), #charset);
				#output->(replace: #token, #newvalue);
			/if;
		/loop;
		
		return: #output;

	/define_tag;

	// [Decode_BHeader] - Decodes email header values using B (binhex) algorithm
	//
	// -Data = (string, the text to be decoded)
	//

	define_tag:'bheader',
		-Priority='Replace',
		-required='data', -type='string',
		-namespace = 'decode';

		if: (#data !>> '=?') || (#data !>> '?b?') || (#data !>> '?=');
			return: #data;
		/if;
		
		local: 'output' = (string_replaceregexp: #data, -find='\\?=\\s+', -replace='?=');
		local: 'output' = (string_replaceregexp: #output, -find='\r\n\\s+', -replace='');
		
		local: 'parse' = (string_findregexp: #output, -find='=\\?([a-zA-Z0-9_-]+?)\\?B\\?(.*?)\\?=');

		if: #parse->size == 0;
			return: #data;
		/if;
		
		loop: -from=1, -to=#parse->size, -by=3;
			local: 'token' = #parse->(get: loop_count);
			local: 'charset' = #parse->(get: loop_count + 1);
			local: 'value' = #parse->(get: loop_count + 2);

			if: (string_validcharset: #charset);
				local: 'newvalue' = (bytes: (decode_base64: #value), #charset);
				#output->(replace: #token, #newvalue);
			/if;
		/loop;
		
		return: #output;

	/define_tag;

?>
