<?LassoScript

// email_pop - POP Mail Downloader for Lasso Professional 8
//
// Quick Start - This type is designed to be used as a high level POP interface.  The basic operation is:
//
// [var: 'pop' = (email_pop: -host='mail.example.com', -username='username', -password='', -apop=true, -log=true)]
// [iterate: $pop, (var: 'msg')]
//   [var: 'msg']
//   [$pop->(delete)]
// [/iterate]
// [output: $pop->(close)]
//
// High-Level Commands - These are the recommended tags for most users
//
// [email_pop:
//   -host='mail.example.com',
//   -port=110, (optional)
//   -username='username', (if a username and password are specifed then
//   -password='password', the type will auto-authorize the POP server)
//   -apop=true, (optional, APOP will be tried before USER/PASS)
//   -log=false, (optional, log all actions to the console)
//   -timeout=15 (optional, timeout for al reads and writes 
// ]
// [email_pop->authorize:
//   -username='username',
//   -password='password',
//   -apop=true, (optional, APOP will be tried before USER/PASS)
// ]
// [email_pop->size] returns the number of messages availble for download
// [email_pop->(get: x)] returns the specified message from the server
// [email_pop->(delete: x)] deletes the specified message from the server
// [email_pop->(close)] closes the POP connection, performing any specified deletes
// [email_pop->(cancel)] closes the POP connection, but does not perform any deletes
//
// The size and get tags allow the iterate code above to be used.  The delete tag
// will delete the last message which was specified for get if no parameter is passed.
//
// Low-Level Commands - These tags allow access to the raw POP commands
//
// [email_pop->(USER: 'username')] sends a username to the server
// [email_pop->(PASS: 'username')] sends a password to the server
// [email_pop->(APOP: 'username','digest')] sends APOP authentication to the server
// [email_pop->(QUIT)] closes the connection to the server
// [email_pop->(RSET)] resets the server, cancelling any deletes
// [email_pop->(STAT] returns the status of the server
// [email_pop->(LIST)] returns a list of index and sizes for all messages on server
// [email_pop->(LIST: x)] returns the index and size for message x.
// [email_pop->(RETR: x)] retrieves message x from the server
// [email_pop->(TOP: x, y)] fetches message x from the server, getting the headers and y lines
// [email_pop->(DELE: x)] deletes message x from the server
// [email_pop->(NOOP)] a null operation which can be used to keep a connection open.
//
// Implementation - These tags are used in the implementation and shouldn't be accessed directly
//
// [email_pop->(pop_cmd:
//   -command='POP command',
//   -timeout=15, (optional timeout for this command)
//   -multi=false (if true a long response to the command (ended with .) is collected
// ]
// [email_pop->(pop_err:
//   -msg='error message',
//   -code=-1, (optional, result code)
//   -cmd='POP command' (optional)
// ]
// [email_pop->(pop_log:
//   -msg='log message',
//   -code=0, (optional, result code)
//   -cmd='POP command' (optional)
// ]
//

define_type: 'pop',
	-namespace = 'email_';
	
	local: 'pop_net' = null;
	local: 'pop_mode' = '';	// init,auth,live,done
	local: 'pop_timeout' = 15;
	local: 'pop_token' = '';
	local: 'pop_index' = 0;
	local: 'pop_err' = (array);
	local: 'pop_res' = (array);
	local: 'pop_log' = false;
	local: 'pop_get' = 'retr';
	local: 'pop_ids' = map;
	
	// Call Back Tags
	
	define_tag: 'onCreate',
			-optional='server',
			-optional='port',
			-optional='username',
			-optional='password',
			-optional='apop',
			-optional='timeout',
			-optional='log',
			-optional='get',
			-optional='host';

		self->'pop_get' = 'init';
		self->'pop_net' = (net);
		self->'pop_log' = (boolean: (local: 'log'));
		
		if: (local_defined: 'get');
			self->'pop_get' = (string: #get);
		else;
			self->'pop_get' = 'uidl';
		/if;
		
		if: (local_defined: 'apop');
			local: '_apop' = (boolean: #apop);
		else;
			local: '_apop' = true;
		/if;

		if: (local_defined: 'server');
			local: '_server' = (string: #server);
			#_server->trim;
		else: (local_defined: 'host');
			local: '_server' = (string: #host);
			#_server->trim;
		/if;
		
		if: (local_defined: 'port');
			local: '_port' = (integer: #port);
		else;
			local: '_port' = 110;
		/if;

		local: '_timeout' = (integer: self->'pop_timeout');
		if: (local_defined: 'timeout');
			#_timeout = (integer: #timeout);
			if: (#_timeout < 5);
				local: '_timeout' = 5;
			/if;
			self->'pop_timeout' = (integer: #_timeout);
		/if;

		self->'pop_net'->(connect: #_server, #_port);

		if: (self->'pop_net'->(wait: #_timeout, net_waitread) == net_waitread);
			local: 'result' = self->'pop_net'->(read: 32768);
			protect;
				self->'pop_token' = (string_findregexp: #result, -find='<[^>]+>')->(get:1);
			/protect;
			self->(pop_log: -msg=#result->(split: '\r\n')->(get:1));
		else;
			self->(pop_err: -msg='The remote server timed out on connect.');
		/if;
		
		if: (local_defined: 'username') && (local_defined: 'password');
			self->(authorize: -username=(local: 'username'), -password=(local: 'password'), -apop=(local: '_apop'));
		/if;
			
	/define_tag;
	

	define_tag: 'onDestroy';

		protect;
			if: (self->'pop_mode' != 'done');
				self->quit;
			/if;
			
			handle: true;
				if: (self->'pop_net'->(isa: 'net'));
					self->'pop_net'->close;
					self->'pop_net' = null;
				/if;
			/handle;
		/protect;
		
	/define_tag;


	define_tag: 'onConvert',
			-required='type';
	
	/define_tag;
	

	// Communication Commands

	define_tag: 'pop_cmd', 
			-required='command',
			-optional='timeout',
			-optional='multi';

		local: 'result' = '';
		if: (self->'pop_net'->(isa: 'net'));
			
			local: '_command' = (string: #command);
			if: !(#_command->(endswith: '\r\n'));
				#_command += '\r\n';
			/if;
			
			local: '_timeout' = (integer: self->'pop_timeout');
			if: (local_defined: 'timeout');
				#_timeout = (integer: #timeout);
			/if;
			
			local: '_multi' = (boolean: (local: 'multi'));
			
			if: (self->'pop_net'->(wait: #_timeout, net_waitwrite) == net_waitwrite);
				local: 'write' = (integer: self->'pop_net'->(write: #_command));
				if: (self->'pop_net'->(wait: #_timeout, net_waitread) == net_waitread);
					local: 'result' = self->'pop_net'->(read: 32768);
					self->(pop_log: -msg=#result->(split: '\r\n')->(get:1), -cmd=#command);
					if: (#_multi == true) && (#result !>> '\r\n.\r\n') && (#result !>> '-ERR');
						while: (#result !>> '\r\n.\r\n');
							if: (loop_count > 10000);
								self->(pop_err: -msg='Error while reading from remote server.', -cmd=#command);
								loop_abort;
							/if;
							local: 'temp' = self->'pop_net'->(read: 32768);
							self->(pop_log: -msg='... reading (' + #temp->size + ' bytes)', -cmd=#command);
							if: (#temp->size == 0);
								loop_abort;
							/if;
							#result += #temp;
						/while;
					/if;
				else;
					self->(pop_err: -msg='The remote server timed out on read.', -cmd=#command);
				/if;
			else;
				self->(pop_err: -msg='The remote server timed out on write.', -cmd=#command);
			/if;
		
		else;
			self->(pop_err: -msg='Lost network connection unexpectedly.', -cmd=#command);
		/if;
		return: #result;
	
	/define_tag;

	define_tag: 'pop_err',
			-required='msg',
			-optional='code',
			-optional='cmd';
		
		if: (local_defined: 'cmd');
			local: '_cmd' = (string: #cmd);
		else;
			local: '_cmd' = '(none)';
		/if;
		
		if: (local_defined: 'code');
			local: '_code' = (integer: #code);
		else;
			local: '_code' = -1;
		/if;
		
		if: (self->'pop_err'->size > 0) && (self->'pop_err'->last == null);
			self->'pop_err'->remove;
		/if;
			
		self->'pop_err'->(insert: #_cmd=#msg);
		self->'pop_res'->(insert: #_cmd=#msg);
		
		if: (self->'pop_log');
			log_detail: 'POP: ' + #_cmd + ' - ' + #msg;
		/if;
		
		fail: #_code, #msg;
		
	/define_tag;
	
	define_tag: 'pop_log',
			-required='msg',
			-optional='code',
			-optional='cmd';

		if: (local_defined: 'cmd');
			local: '_cmd' = (string: #cmd);
		else;
			local: '_cmd' = '(none)';
		/if;
		
		if: (local_defined: 'code');
			local: '_code' = (integer: #code);
		else;
			local: '_code' = -1;
		/if;
		
		if: (self->'pop_err'->size > 0) && (self->'pop_err'->last != null);
			self->'pop_err'->(insert: null);
		/if;
			
		self->'pop_res'->(insert: #_cmd=#msg);
		
		if: (self->'pop_log');
			log_detail: 'POP: ' + #_cmd + ' - ' + #msg;
		/if;
		
		error_seterrormessage: (error_noerror);
		error_seterrorcode: (error_noerror: -errorcode);
		
	/define_tag;
	
	
	// Low Level Commands

	define_tag: 'user',
			-optional='username';
		if: (local_defined: 'username');
			return: self->(pop_cmd: 'USER ' + #username);
		else;
			self->(pop_err: -msg='No username specified.',-cmd='USER');
		/if;
	/define_tag;

	define_tag: 'pass',
			-optional='password';
		if: (local_defined: 'password');
			return: self->(pop_cmd: 'PASS ' + #password);
		else;
			self->(pop_err: -msg='No password specified.',-cmd='PASS');
		/if;
	/define_tag;

	define_tag: 'apop',
			-required='username',
			-required='password';
		if: (self->'pop_token' >> '<');
			local: 'digest' = (encrypt_md5: self->'pop_token' + #password);
			return: self->(pop_cmd: 'APOP ' + #username + ' ' + #digest);
		else;
			self->(pop_err: -msg='No token found.',-cmd='APOP');
		/if;
	/define_tag;

	define_tag: 'quit';
		self->'pop_mode' = 'done';
		return: self->(pop_cmd: 'QUIT')->trim &;
	/define_tag;

	define_tag: 'rset';
		return: self->(pop_cmd: 'RSET')->trim &;
	/define_tag;

	define_tag: 'stat';
		return: self->(pop_cmd: 'STAT')->trim &;
	/define_tag;

	define_tag: 'list',
			-optional='msg';
		if: (local_defined: 'msg');
			return: self->(pop_cmd: 'LIST ' + (integer: #msg))->trim &;
		else;
			return: self->(pop_cmd: 'LIST', -multi=true)->trim &;
		/if;
	/define_tag;

	define_tag: 'uidl',
			-optional='msg';
		if: (local_defined: 'msg');
			local: 'result' = 
			return: self->(pop_cmd: 'UIDL ' + (integer: #msg))->trim &;
		else;
			return: self->(pop_cmd: 'UIDL', -multi=true)->trim &;
		/if;
	/define_tag;

	define_tag: 'retr',
			-required='msg';
		return: self->(pop_cmd: 'RETR ' + (integer: #msg), -multi=true)->trim &;
	/define_tag;

	define_tag: 'top',
			-required='msg',
			-optional='size';
		return: self->(pop_cmd: 'TOP ' + (integer: #msg) + ' ' + (integer: (local: 'size')), -multi=true)->trim &;
	/define_tag;

	define_tag: 'dele',
			-required='msg';
		return: self->(pop_cmd: 'DELE ' + (integer: #msg))->trim &;
	/define_tag;

	define_tag: 'noop';
		return: self->(pop_cmd: 'NOOP')->trim &;
	/define_tag;


	// High Level Commands
	
	define_tag: 'authorize',
			-required='username',
			-required='password',
			-optional='apop';

		local: 'done' = false;
		if: (local_defined: 'apop');
			local: '_apop' = (boolean: #apop);
		else;
			local: '_apop' = true;
		/if;
		if: (#_apop && (self->'pop_token' >> '<'));
			self->(apop: -username=#username, -password=#password);
			local: 'done' = (self->'pop_res'->last->second >> '+OK');
		/if;
		if: (#done == false);
			self->(user: #username);
			self->(pass: #password);
		/if;
		if: (self->'pop_res'->last->second >> '+OK');
			self->'pop_mode' = 'ready';
		else;
			self->(pop_err: -msg='Could not authenticate user',-cmd='PASS');
		/if;
		self->'pop_ids' = map;
		self->'pop_index' = 0;
	/define_tag;

	define_tag: 'size';
		protect;
			return: (integer: (string_findregexp: self->(pop_cmd: 'STAT'), -find='[0-9]+')->(get:1));
		/protect;
		return: 0;
	/define_tag;
	
	define_tag: 'get',
			-optional='msg';
		if: (local_defined: 'msg') && !(string_isdigit: #msg);
			if: (self->'pop_ids' >> #msg);
				local: 'dex' = self->'pop_ids'->(find: #msg);
			else;
  				fail_if: (local_defined: 'msg'), -9956, 'The tag [Email_POP->' + tag_name + '] expected an integer position.';
			/if;
		else;
			local: 'dex' = (integer: (local: 'msg'));
		/if;
		if: (#dex == 0);
			local: 'dex' = (integer: self->'pop_index');
		/if;
		self->'pop_index' = #dex;
		if: #dex > 0;
			if: (self->'pop_get' >> 'retr');
				return: self->(retrieve: #dex);
			else: (self->'pop_get' >> 'dele');
				return: self->(delete: #dex);
			else: (self->'pop_get' >> 'head');
				return: self->(headers: #dex);
			else: (self->'pop_get' >> 'top');
				local: 'size' = (integer: (string_replaceregexp: self->'pop_get', -find='.*(\\d+).*', -replace='\\1'));
				return: self->(retrieve: #dex, #size);
			else;
				return: self->(uniqueid: #dex);
			/if;
		/if;
		return: '';
	/define_tag;
	
	define_tag: 'retrieve',
			-optional='msg',
			-optional='size';
  		fail_if: (local_defined: 'size') && !(string_isdigit: #size), -9956, 'The tag [Email_POP->' + tag_name + '] expected an integer size.';
		if: (local_defined: 'msg') && !(string_isdigit: #msg);
			if: (self->'pop_ids' >> #msg);
				local: 'dex' = self->'pop_ids'->(find: #msg);
			else;
  				fail_if: (local_defined: 'msg'), -9956, 'The tag [Email_POP->' + tag_name + '] expected an integer position.';
			/if;
		else;
			local: 'dex' = (integer: (local: 'msg'));
		/if;
		if: (#dex == 0);
			local: 'dex' = (integer: self->'pop_index');
		/if;
		self->'pop_index' = #dex;
		if: #dex > 0;
			if: (local_defined: 'size');
				local: 'result' = self->(top: #dex, (integer: #size));
			else;
				local: 'result' = self->(retr: #dex);
			/if;
			fail_if: (#result->(beginswith: '-ERR')), -9956, 'The tag [Email_POP->' + tag_name + '] did not find a message at provided position (' + #dex ').';
			local: 'index' = #result->(find: '\r\n');
			if: (#index > 0);
				local: 'result' = (_substring: #result, #index + 2);
			/if;
			#result->trim;
			#result->(removetrailing: '\r\n.');
			return: #result;
		/if;
		return: '';
	/define_tag;
	
	define_tag: 'headers',
			-optional='msg';
		if: (local_defined: 'msg') && !(string_isdigit: #msg);
			if: (self->'pop_ids' >> #msg);
				local: 'dex' = self->'pop_ids'->(find: #msg);
			else;
  				fail_if: (local_defined: 'msg'), -9956, 'The tag [Email_POP->' + tag_name + '] expected an integer position.';
			/if;
		else;
			local: 'dex' = (integer: (local: 'msg'));
		/if;
		if: (#dex == 0);
			local: 'dex' = (integer: self->'pop_index');
		/if;
		self->'pop_index' = #dex;
		if: #dex > 0;
			local: 'result' = self->(top: #dex, 0);
			fail_if: (#result->(beginswith: '-ERR')), -9956, 'The tag [Email_POP->' + tag_name + '] did not find a message at provided position (' + #dex ').';
			local: 'index' = #result->(find: '\r\n');
			if: (#index > 0);
				local: 'result' = (_substring: #result, #index + 2);
			/if;
			#result->trim;
			#result->(removetrailing: '\r\n.');
			return: #result;
		/if;
		return: '';
	/define_tag;
	
	define_tag: 'delete',
			-optional='msg';
		if: (local_defined: 'msg') && !(string_isdigit: #msg);
			if: (self->'pop_ids' >> #msg);
				local: 'dex' = self->'pop_ids'->(find: #msg);
			else;
  				fail_if: (local_defined: 'msg'), -9956, 'The tag [Email_POP->' + tag_name + '] expected an integer position.';
			/if;
		else;
			local: 'dex' = (integer: (local: 'msg'));
		/if;
		if: (#dex == 0);
			local: 'dex' = (integer: self->'pop_index');
		/if;
		self->'pop_index' = 0;
		if: #dex > 0;
			local: 'result' = self->(dele: #dex);
			fail_if: (#result->(beginswith: '-ERR')), -9956, 'The tag [Email_POP->' + tag_name + '] did not find a message at provided position (' + #dex ').';
		/if;
	/define_tag;
	
	define_tag: 'uniqueid',
			-optional='msg';
  		fail_if: (local_defined: 'msg') && !(string_isdigit: #msg), -9956, 'The tag [Email_POP->' + tag_name + '] expected an integer position.';
		local: 'dex' = (integer: (local: 'msg'));
		if: (#dex == 0);
			local: 'dex' = (integer: self->'pop_index');
		/if;
		self->'pop_index' = #dex;
		if: #dex > 0;
			local: 'result' = self->(uidl: #dex);
			fail_if: (#result->(beginswith: '-ERR')), -9956, 'The tag [Email_POP->' + tag_name + '] did not find a message at provided position (' + #dex ').';
			local: 'array' = #result->(split: ' ');
			if: (#array->size > 2);
				local: 'id' = #array->(get: 3);
				if: !(string_isdigit: #id);
					self->'pop_ids'->(insert: #id = #dex);
				/if;
				return: @#id;
			/if;
		/if;
		return: '';
	/define_tag;
	
	define_tag: 'close';
		self->'pop_ids' = map;
		self->'pop_index' = 0;
		self->(quit);
	/define_tag;
	
	define_tag: 'cancel';
		self->'pop_ids' = map;
		self->'pop_index' = 0;
		self->(rset);
		self->(quit);
	/define_tag;
	
	define_tag: 'results';
		return: self->'pop_res';
	/define_tag;
	
	define_tag: 'errors';
		return: self->'pop_err';
	/define_tag;
	
	define_tag: 'lasterror';
		return: self->'pop_err'->last;
	/define_tag;
			
	define_tag: 'data';
		return: self->'pop_token';
	/define_tag;
	
/define_type;

	
	// email_parse
	
define_type: 'parse',
	-namespace = 'email_';

	local: 'parse_msg';
	local: 'parse_rawhdrs';
	local: 'parse_hdrs';
	local: 'parse_body';
	local: 'parse_content_type';
	local: 'parse_content_transfer_encoding';
	local: 'parse_content_disposition';
	local: 'parse_charset';
	local: 'parse_boundary';
	local: 'parse_parts';
	local: 'parse_mode';

	define_tag: 'onCreate',
			-required='msg';
		
		fail_if: !(#msg->(isa: 'string') || #msg->(isa: 'bytes')), -9956, '[Email_Parse] expects a string or bytes parameter.';
		fail_if: (#msg == ''), -9956, '[Email_Parse] expects a non-empty email message to parse.';
		
		self->'parse_msg' = (local: 'msg');
		self->'parse_msg'->(replace: '\r\n', '\r');
		self->'parse_msg'->(replace: '\n', '\r');
		self->'parse_msg'->(replace: '\r', '\r\n');

		local: 'index' = self->'parse_msg'->(find: '\r\n\r\n');
		if: #index > 0;
			self->'parse_rawhdrs' = (_substring: self->'parse_msg', 1, #index);
		else;
			self->'parse_rawhdrs' = self->'parse_msg';
		/if;
		self->'parse_rawhdrs'->trim;
		self->'parse_hdrs' = (_chop_mimes: self->'parse_rawhdrs');
		
		self->'parse_body' = (_substring: self->'parse_msg', #index);
		self->'parse_body'->trim;
		self->'parse_body'->(removetrailing: '.');
		self->'parse_body'->trim;

		self->'parse_content_type' = self->(header: 'content-type');
		self->'parse_content_transfer_encoding' = self->(header: 'content-transfer-encoding');
		self->'parse_content_disposition' = self->(header: 'content-disposition');

		self->'parse_mode' = self->(content_type: 'type');

		if: (self->'parse_mode' >> 'multipart');
		
			// Decompose Multi-Part Message
			self->'parse_boundary' = '--' + self->boundary;
			if: (self->'parse_boundary' != '--') && (self->'parse_body' >> self->'parse_boundary');
				self->'parse_parts' = self->'parse_body'->(split: self->'parse_boundary');
				loop: -from=self->'parse_parts'->size, -to=1, -by=-1;
					if: (self->'parse_parts'->(get: loop_count) == '');
						self->'parse_parts'->(remove: loop_count);
					else: (self->'parse_parts'->(get: loop_count) == '--');
						self->'parse_parts'->(remove: loop_count);
					/if;
				/loop;
			else;
				self->'parse_parts' = (array: self->'parse_body');
			/if;
			
		else;
		
			// Simple Message
			self->'parse_charset' = self->charset;
			self->'parse_parts' = null;
		
		/if;
	/define_tag;
	
	define_tag: 'onConvert';
		local: 'output' = '';
		local: 'temp' = '';
		(#temp := self->date) ? #output += 'Date: ' + #temp + '\r\n';
		(#temp := self->from) ? #output += 'From: ' + (email_safeemail: #temp) + '\r\n';
		(#temp := self->to) ? #output += 'To: ' + (email_safeemail: #temp) + '\r\n';
		(#temp := self->cc) ? #output += 'Cc: ' + (email_safeemail: #temp) + '\r\n';
		(#temp := self->subject) ? #output += 'Subject: ' + #temp + '\r\n';
		(#temp := self->content_type) ? #output += 'Content-Type: ' + #temp + '\r\n';
		(#temp := self->content_transfer_encoding) ? #output += 'Content-Transfer-Encoding: ' + #temp + '\r\n';
		(#temp := self->content_disposition) ? #output += 'Content-Disposition: ' + #temp + '\r\n';
		(#temp := self->'parse_parts'->size) ? #output += 'Parts: ' + #temp + '\r\n';
		(#output != '') ? #output += '\r\n';
		#output += self->body;
		return: #output;
	/define_tag;
	
	define_tag: 'raw';
		return: self->'parse_msg';
	/define_tag;
			
	define_tag: 'data';
		return: self->'parse_msg';
	/define_tag;
			
	define_tag: 'headers';
		return: self->'parse_hdrs';
	/define_tag;
	
	define_tag: 'rawheaders';
		return: self->'parse_rawhdrs';
	/define_tag;
	
	define_tag: 'header',
			-required='name',
			-optional='join',
			-optional='comment',
			-optional='extract',
			-optional='safeemail',
			-optional='nodecode';
			
		local: 'headers' = self->'parse_hdrs'->(find: #name);
		local: 'array' = (array);

		iterate: #headers, (local: 'temp');
			local: 'output' = #temp->second;
			if: !(local_defined: 'nodecode');
				if: (#output >> '=?') && (#output >> '?q?') && (#output >> '?=');
					local: 'output' = (decode_qheader: #output);
				/if;
				if: (#output >> '=?') && (#output >> '?b?') && (#output >> '?=');
					local: 'output' = (decode_bheader: #output);
				/if;
			/if;
			if: (local_defined: 'comment');
				local: 'output' = (email_extract: #output, -comment);
			else: (local_defined: 'extract');
				local: 'output' = (email_extract: #output);
			else: (local_defined: 'safeemail');
				local: 'output' = (email_safeemail: #output);
			/if;
			#array->(insert: #output);
		/iterate;
		
		if: (#array->size == 0);
			return: null;
		else: (#array->size == 1);
			return: #array->first;
		else: (local_defined: 'join');
			return: #array->(join: #join);
		else;
			return: #array;
		/if;
		
	/define_tag;
	
	define_tag: 'content_type',
			-optional='fld';
		if: !(local_defined: 'fld') || (#fld == '');
			return: self->'parse_content_type';
		else: (#fld == 'type');
			local: 'temp' = (string_findregexp: self->'parse_content_type', -find='[0-9a-zA-Z_\\-]+(?=/)');
			(#temp->size > 0) ? return: #temp->first;
		else: (#fld == 'subtype');
			local: 'temp' = (string_findregexp: self->'parse_content_type', -find='(?<=/)[0-9a-zA-Z_\\-]+');
			(#temp->size > 0) ? return: #temp->first;
		else;
			local: 'temp' = (string_findregexp: self->'parse_content_type', -find='(?<=' + #fld + '=")[^"]+(?=")');
			(#temp->size > 0) ? return: #temp->first;
			local: 'temp' = (string_findregexp: self->'parse_content_type', -find='(?<=' + #fld + '=)[^;\r\n\t ]+');
			(#temp->size > 0) ? return: #temp->first;
		/if;
		return: '';
	/define_tag;
	
	define_tag: 'content_transfer_encoding';
		return: self->'parse_content_transfer_encoding';
	/define_tag;
	
	define_tag: 'content_disposition';
		return: self->'parse_content_disposition';
	/define_tag;
	
	define_tag: 'mode';
		return: self->(content_type: 'type');
	/define_tag;
	
	define_tag: 'boundary';
		return: self->(content_type: 'boundary');
	/define_tag;
	
	define_tag: 'charset';
		return: self->(content_type: 'charset');
	/define_tag;
	
	define_tag: 'to',
			-optional='join',
			-optional='comment',
			-optional='extract',
			-optional='safeemail';
		if: !(local_defined: 'join');
			local: 'join' = ', ';
		/if;
		if: (local_defined: 'safeemail');
			return: self->(header: 'to', -join=#join, -safeemail);
		else: (local_defined: 'comment');
			return: self->(header: 'to', -join=#join, -comment);
		else: (local_defined: 'extract');
			return: self->(header: 'to', -join=#join, -extract);
		else;
			return: self->(header: 'to', -join=#join);
		/if;
	/define_tag;
	
	define_tag: 'from',
			-optional='join',
			-optional='comment',
			-optional='extract',
			-optional='safeemail';
		if: !(local_defined: 'join');
			local: 'join' = ', ';
		/if;
		if: (local_defined: 'safeemail');
			return: self->(header: 'from', -join=#join, -safeemail);
		else: (local_defined: 'comment');
			return: self->(header: 'from', -join=#join, -comment);
		else: (local_defined: 'extract');
			return: self->(header: 'from', -join=#join, -extract);
		else;
			return: self->(header: 'from', -join=#join);
		/if;
	/define_tag;
	
	define_tag: 'subject';
		return: self->(header: 'subject', -join=', ');
	/define_tag;
	
	define_tag: 'cc',
			-optional='join',
			-optional='comment',
			-optional='extract',
			-optional='safeemail';
		if: !(local_defined: 'join');
			local: 'join' = ', ';
		/if;
		if: (local_defined: 'safeemail');
			return: self->(header: 'cc', -join=#join, -safeemail);
		else: (local_defined: 'comment');
			return: self->(header: 'cc', -join=#join, -comment);
		else: (local_defined: 'extract');
			return: self->(header: 'cc', -join=#join, -extract);
		else;
			return: self->(header: 'cc', -join=#join);
		/if;
	/define_tag;
	
	define_tag: 'bcc',
			-optional='join',
			-optional='comment',
			-optional='extract',
			-optional='safeemail';
		if: !(local_defined: 'join');
			local: 'join' = ', ';
		/if;
		if: (local_defined: 'safeemail');
			return: self->(header: 'bcc', -join=#join, -safeemail);
		else: (local_defined: 'comment');
			return: self->(header: 'bcc', -join=#join, -comment);
		else: (local_defined: 'extract');
			return: self->(header: 'bcc', -join=#join, -extract);
		else;
			return: self->(header: 'bcc', -join=#join);
		/if;
	/define_tag;
	
	define_tag: 'date';
		local: 'formats' = (array: 
			'%a %-d %b %Y %-H:%-M:%-S %z',
			'%a , %b %-d %Y %H:%M:%S %z',
			'%-d %b %Y %T %z',
			'%_d %b %Y %T %z',
			'%a, %_d %b %Y %H:%M:%S %z',
			'%-d-%b-%Y %T %z',
			'%_d-%b-%Y %T %z');
		local: 'date' = self->(header: 'date');
		iterate: #formats, (local: 'format');
			protect;
				local: 'output' = (date_gmttolocal: (date: #date, -format=#format));
				if: (#output != '');
					return: @#output;
				/if;
			/protect;
		/iterate;
		protect;
			local: 'output' = (date_gmttolocal: (date: #date));
			if: (#output != '');
				return: @#output;
			/if;
		/protect;
		return: @#date;
	/define_tag;
	
	define_tag: 'body',
			-optional='type';
		if: (self->'parse_parts'->size == 0);
		
			local: 'output' = self->'parse_body';
			if: (self->'parse_content_transfer_encoding' >> 'quoted-printable');
				local: 'output' = (decode_quotedprintable: (string: #output), self->'parse_charset');
			else: (self->'parse_content_transfer_encoding' >> 'base64');
				#output->trim;
				(#output >> '\r\n') ? #output->(replace: '\r\n', '');
				(#output >> '\r') ? #output->(replace: '\r', '');
				(#output >> '\r') ? #output->(replace: '\n', '');
				(#output >> '\r') ? #output->(replace: '\t', '');
				(#output >> '\r') ? #output->(replace: ' ', '');
				local: 'output' = (decode_base64: #output);
			else: #output->(isa: 'bytes') && (string_validcharset: self->'parse_charset');
				local: 'output' = #output->(exportstring: self->'parse_charset');
			/if;
			if: (self->'parse_content_type' >> 'text/enriched');
				// Try to convert text/enriched to HTML
				#output = (string_replaceregexp: #output, -find='<(/?)bold>', -replace='<\\1b>');
				#output = (string_replaceregexp: #output, -find='<(/?)italic>', -replace='<\\1i>');
				#output = (string_replaceregexp: #output, -find='<(/?)excerpt>', -replace='<\\1quote>');
				#output = (string_replaceregexp: #output, -find='<bigger>', -replace='<font size="+1">');
				#output = (string_replaceregexp: #output, -find='<smaller>', -replace='<font size="-1">');
				#output = (string_replaceregexp: #output, -find='<fontfamily><param>(.*?)</param>', -replace='<font face="\\1">');
				#output = (string_replaceregexp: #output, -find='<color><param>..(..),..(..),..(..)</param>', -replace='<font color="#\\1\\2\\3">');
				#output = (string_replaceregexp: #output, -find='</(?:smaller|bigger|fontfamily|color)>', -replace='</font>');
			/if;
			return: @#output;
			
		else: (self->'parse_parts'->size == 1);
			return: (email_parse: -msg=self->'parse_parts'->(get: 1))->(body: (local: 'type'));
		else;
			local: 'types' = (array);
			iterate: self, (local: 'temp');
				#types->(insert: #temp->content_type);
			/iterate;
			if: (local_defined: 'type') && (#type != '');
				iterate: #types, (local: 'temp');
					if: (#temp >> #type);
						return: (email_parse: -msg=self->'parse_parts'->(get: loop_count))->(body: (local: 'type'));
					/if;
				/iterate;
			else;
				iterate: #types, (local: 'temp');
					if: (#temp >> 'text/plain');
						return: (email_parse: -msg=self->'parse_parts'->(get: loop_count))->(body: (local: 'type'));
					/if;
				/iterate;
				iterate: #types, (local: 'temp');
					if: (#temp >> 'text/html');
						return: (email_parse: -msg=self->'parse_parts'->(get: loop_count))->(body: (local: 'type'));
					/if;
				/iterate;
				iterate: #types, (local: 'temp');
					if: (#temp >> 'text/');
						return: (email_parse: -msg=self->'parse_parts'->(get: loop_count))->(body: (local: 'type'));
					/if;
				/iterate;
			/if;
			return: (email_parse: -msg=self->'parse_parts'->(get: 1))->(body: (local: 'type'));
		/if;
	/define_tag;
	
	define_tag: 'size';
		return: self->'parse_parts'->size;
	/define_tag;
	
	define_tag: 'get',
			-required='count';
		return: (email_parse: -msg=self->'parse_parts'->(get: #count));
	/define_tag;

	define_tag: 'first';
		return: (email_parse: -msg=self->'parse_parts'->(get: 1));
	/define_tag;

	define_tag: 'second';
		return: (email_parse: -msg=self->'parse_parts'->(get: 2));
	/define_tag;

	define_tag: 'last';
		return: (email_parse: -msg=self->'parse_parts'->(get: self->'parse_parts'->size));
	/define_tag;

/define_type;


// [Email_SafeEmail]
//
// First parameter is a string containing the contents of a To/From header.
//
// This tag returns a 'safe' version of the user's address using their
// friendly name or the first part of their email address.
//

define_tag: 'SafeEmail',
		-Required='data',
		-Priority='Replace',
		-namespace = 'email_';
	local: 'output' = (Email_Extract: #data, -comment);
	if: (#output == '');
		local: 'output' = (Email_Extract: #data);
	/if;
	if: (#output >> '@');
		local: 'index' = #output->(find: '@');
		local: 'output' = (_substring: #output, 1, #index - 1);
	/if;
	#output->trim;
	#output->(removeleading: '"');
	#output->(removetrailing: '"');
	return: #output;
/define_tag;


// [Email_Extract]
//
// First parameter is a string containing the contents of a header.
//
// This tag returns a 'safe' version of the user's address using their
// friendly name or the first part of their email address.
//

define_tag: 'Extract',
		-Required='data',
		-Optional='comment',
		-Priority='Replace',
		-namespace = 'email_';
	if: (local_defined: 'comment');
		protect;
			local: 'output' = (string_findregexp: #data, -find='\\(([^\\)]+)\\)')->(get:2);
			#output->trim;
			#output->(removeleading: '(');
			#output->(removetrailing: ')');
			local: 'output' = (string_replaceregexp: #output, -find='\\s+', -replace=' ');
			return: #output;
		/protect;
		local: 'output' = (string_replaceregexp: #data, -find='<[^>]+>', -replace='');
		#output->trim;
		local: 'output' = (string_replaceregexp: #output, -find='\\s+', -replace=' ');
		return: #output;
	else;
		protect;
			local: 'output' = (string_findregexp: #data, -find='<([^>]+)>')->(get:2);
			#output->trim;
			#output->(removeleading: '<');
			#output->(removetrailing: '>');
			local: 'output' = (string_replaceregexp: #output, -find='\\s+', -replace=' ');
			return: #output;
		/protect;
		local: 'output' = (string_replaceregexp: #data, -find='\\([^\\)]+\\)', -replace='');
		#output->trim;
		local: 'output' = (string_replaceregexp: #output, -find='\\s+', -replace=' ');
		return: #output;
	/if;
/define_tag;

define_tag: '_substring', -required='data', -optional='start', -optional='length';
	if: #data->(isa: 'bytes');
		if: ((local: 'start') !== null) && (#start > 0);
			if: ((local: 'length') !== null) && (#length > 0);
				return: #data->(getrange: #start, #length);
			else;
				return: #data->(getrange: #start, (1 + #data->size - #start));
			/if;
		/if;
	else: #data->(isa: 'string');
		if: ((local: 'start') !== null) && (#start > 0);
			if: ((local: 'length') !== null) && (#length > 0);
				return: #data->(substring: #start, #length);
			else;
				return: #data->(substring: #start);
			/if;
		/if;
	else;
		return: (_substring: (string: #data), -start=(local: 'start'), -length=(local: 'length'));
	/if;
	return: #data;
/define_tag;

?>
